"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Publisher = void 0;
var tslib_1 = require("tslib");
var analytics_core_1 = require("@segment/analytics-core");
var node_fetch_1 = tslib_1.__importDefault(require("node-fetch"));
var create_url_1 = require("../../lib/create-url");
var extract_promise_parts_1 = require("../../lib/extract-promise-parts");
var context_batch_1 = require("./context-batch");
function sleep(timeoutInMs) {
    return new Promise(function (resolve) { return setTimeout(resolve, timeoutInMs); });
}
function noop() { }
/**
 * The Publisher is responsible for batching events and sending them to the Segment API.
 */
var Publisher = /** @class */ (function () {
    function Publisher(_a) {
        var host = _a.host, path = _a.path, maxAttempts = _a.maxAttempts, maxEventsInBatch = _a.maxEventsInBatch, maxWaitTimeInMs = _a.maxWaitTimeInMs, writeKey = _a.writeKey;
        this._maxAttempts = Math.max(maxAttempts, 1);
        this._maxEventsInBatch = Math.max(maxEventsInBatch, 1);
        this._maxWaitTimeInMs = maxWaitTimeInMs;
        this._auth = Buffer.from("".concat(writeKey, ":")).toString('base64');
        this._url = (0, create_url_1.tryCreateFormattedUrl)(host !== null && host !== void 0 ? host : 'https://api.segment.io', path !== null && path !== void 0 ? path : '/v1/batch');
    }
    Publisher.prototype.createBatch = function () {
        var _this = this;
        this.pendingFlushTimeout && clearTimeout(this.pendingFlushTimeout);
        var batch = new context_batch_1.ContextBatch(this._maxEventsInBatch);
        this.batch = batch;
        this.pendingFlushTimeout = setTimeout(function () {
            if (batch === _this.batch) {
                _this.batch = undefined;
            }
            _this.pendingFlushTimeout = undefined;
            if (batch.length) {
                _this.send(batch).catch(noop);
            }
        }, this._maxWaitTimeInMs);
        return batch;
    };
    Publisher.prototype.clearBatch = function () {
        this.pendingFlushTimeout && clearTimeout(this.pendingFlushTimeout);
        this.batch = undefined;
    };
    /**
     * Enqueues the context for future delivery.
     * @param ctx - Context containing a Segment event.
     * @returns a promise that resolves with the context after the event has been delivered.
     */
    Publisher.prototype.enqueue = function (ctx) {
        var _a;
        var batch = (_a = this.batch) !== null && _a !== void 0 ? _a : this.createBatch();
        var _b = (0, extract_promise_parts_1.extractPromiseParts)(), ctxPromise = _b.promise, resolve = _b.resolve;
        var pendingItem = {
            context: ctx,
            resolver: resolve,
        };
        /*
          The following logic ensures that a batch is never orphaned,
          and is always sent before a new batch is created.
    
          Add an event to the existing batch.
            Success: Check if batch is full and send if it is.
            Failure: Assume event is too big to fit in current batch - send existing batch.
              Add an event to the new batch.
                Success: Check if batch is full and send if it is.
                Failure: Event exceeds maximum size (it will never fit), fail the event.
        */
        if (batch.tryAdd(pendingItem)) {
            if (batch.length === this._maxEventsInBatch) {
                this.send(batch).catch(noop);
                this.clearBatch();
            }
            return ctxPromise;
        }
        else if (batch.length) {
            this.send(batch).catch(noop);
            this.clearBatch();
        }
        var fallbackBatch = this.createBatch();
        if (!fallbackBatch.tryAdd(pendingItem)) {
            ctx.setFailedDelivery({
                reason: new Error("Event exceeds maximum event size of 32 kb"),
            });
            return Promise.resolve(ctx);
        }
        else {
            if (fallbackBatch.length === this._maxEventsInBatch) {
                this.send(fallbackBatch).catch(noop);
                this.clearBatch();
            }
            return ctxPromise;
        }
    };
    Publisher.prototype.send = function (batch) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var events, payload, maxAttempts, currentAttempt, failureReason, response, err_1;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        events = batch.getEvents();
                        payload = JSON.stringify({ batch: events });
                        maxAttempts = this._maxAttempts;
                        currentAttempt = 0;
                        _a.label = 1;
                    case 1:
                        if (!(currentAttempt < maxAttempts)) return [3 /*break*/, 7];
                        currentAttempt++;
                        failureReason = void 0;
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, 4, , 5]);
                        return [4 /*yield*/, (0, node_fetch_1.default)(this._url, {
                                method: 'POST',
                                headers: {
                                    'Content-Type': 'application/json',
                                    Authorization: "Basic ".concat(this._auth),
                                    'User-Agent': 'analytics-node-next/latest',
                                },
                                body: payload,
                            })];
                    case 3:
                        response = _a.sent();
                        if (response.ok) {
                            // Successfully sent events, so exit!
                            batch.resolveEvents();
                            return [2 /*return*/];
                        }
                        else if (response.status === 400) {
                            // https://segment.com/docs/connections/sources/catalog/libraries/server/http-api/#max-request-size
                            // Request either malformed or size exceeded - don't retry.
                            resolveFailedBatch(batch, new Error("[".concat(response.status, "] ").concat(response.statusText)));
                            return [2 /*return*/];
                        }
                        else {
                            // Treat other errors as transient and retry.
                            failureReason = new Error("[".concat(response.status, "] ").concat(response.statusText));
                        }
                        return [3 /*break*/, 5];
                    case 4:
                        err_1 = _a.sent();
                        // Network errors get thrown, retry them.
                        failureReason = err_1;
                        return [3 /*break*/, 5];
                    case 5:
                        // Final attempt failed, update context and resolve events.
                        if (currentAttempt === maxAttempts) {
                            resolveFailedBatch(batch, failureReason);
                            return [2 /*return*/];
                        }
                        // Retry after attempt-based backoff.
                        return [4 /*yield*/, sleep((0, analytics_core_1.backoff)({
                                attempt: currentAttempt,
                                minTimeout: 25,
                                maxTimeout: 1000,
                            }))];
                    case 6:
                        // Retry after attempt-based backoff.
                        _a.sent();
                        return [3 /*break*/, 1];
                    case 7: return [2 /*return*/];
                }
            });
        });
    };
    return Publisher;
}());
exports.Publisher = Publisher;
function resolveFailedBatch(batch, reason) {
    batch.getContexts().forEach(function (ctx) { return ctx.setFailedDelivery({ reason: reason }); });
    batch.resolveEvents();
}
//# sourceMappingURL=publisher.js.map